! EOS is that for ideal gas.
! Temperature dependence of Cp is calculated by a least square fitted
! polynomial. The fitted data were Cp profile based on VIRA temperature profile
! by the use of SW1996 EOS.
module thermo_prop_co2_viracplsf

  use vtype_module

  implicit none

  private

  real(dp), save :: ConvThreshold = 1.0d-11

  data ConvThreshold /1.0d-11/

  real(dp), parameter :: UnivGasConst = 8.314510d0  ! J mole-1 K-1
  real(dp), parameter :: MolWeight    = 44.0098d-3  ! kg mole-1
  real(dp), parameter :: GasConst     = UnivGasConst / MolWeight ! J kg-1 K-1

  integer , parameter :: NCoef = 3
  real(dp), save      :: a_CPCoef(NCoef)
  real(dp), parameter :: Temp1 = 450.0d0
  real(dp), parameter :: Temp2 = 300.0d0

!!$  data a_CPCoef / 484.4d0, 1.392d0, -6.104d-4 /
  data a_CPCoef / 987.3d0, 252.8d0, -54.94d0 /


  public :: ThermoPropInit
  public :: ThermoPropInqMolGasConst
  public :: ThermoPropInqMolWeight
  public :: ThermoPropCompFact
  public :: ThermoPropDens
  public :: ThermoPropTemp
  public :: ThermoPropEnthalpy
  public :: ThermoPropCp
  public :: ThermoPropDPressDDensConstTemp
  public :: ThermoPropDPressDTempConstDens
  public :: ThermoPropThermoExpCoef
  public :: ThermoPropAdLapRateDTDP
  public :: ThermoPropDensTempFromPressEnthalpy

  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------
  ! set parameters
  !----------------------------------------------------------------------------

  subroutine ThermoPropInit(     &
    & ConvThresholdSet, ConstCp  &
    & )

    real(dp), intent(in), optional :: ConvThresholdSet
    real(dp), intent(in), optional :: ConstCp


    if ( present( ConvThresholdSet ) ) then
      ConvThreshold = ConvThresholdSet
    end if

    if ( present( ConstCp ) ) then
      a_CPCoef(1)       = ConstCp
      a_CPCoef(2:NCoef) = 0.0d0
    end if


  end subroutine ThermoPropInit

  !----------------------------------------------------------------------------
  ! inquire molar gas constant
  !----------------------------------------------------------------------------

  function ThermoPropInqMolGasConst() &
    & result( Val )

    real(dp) :: Val

    Val = UnivGasConst

  end function ThermoPropInqMolGasConst

  !----------------------------------------------------------------------------
  ! inquire molecular weight
  !----------------------------------------------------------------------------

  function ThermoPropInqMolWeight() &
    & result( Val )

    real(dp) :: Val

    Val = MolWeight

  end function ThermoPropInqMolWeight

  !----------------------------------------------------------------------------
  ! compressibility factor
  !----------------------------------------------------------------------------

  function ThermoPropCompFact( Dens, Temp ) &
    & result( CompFact )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: CompFact


    ! Local variables
    !


    CompFact = 1.0d0


  end function ThermoPropCompFact

  !----------------------------------------------------------------------------
  ! density
  !----------------------------------------------------------------------------

  function ThermoPropDens( Press, Temp ) &
    & result( Dens )

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Temp
    real(dp)             :: Dens


    ! Local variables
    !


    Dens = Press / ( GasConst * Temp )


  end function ThermoPropDens

  !----------------------------------------------------------------------------
  ! temperature
  !----------------------------------------------------------------------------

  function ThermoPropTemp( Press, Dens ) &
    & result( Temp )

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Dens
    real(dp)             :: Temp


    ! Local variables
    !


    Temp = Press / ( Dens * GasConst )


  end function ThermoPropTemp

  !----------------------------------------------------------------------------
  ! enthalpy
  !----------------------------------------------------------------------------

  function ThermoPropEnthalpy( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    integer :: i


    Val = 0.0d0
    do i = 1, NCoef
      Val = Val + a_CPCoef(i) / real( i ) * ( ( Temp - Temp1 ) / Temp2 )**( i ) * Temp2
    end do


  end function ThermoPropEnthalpy

  !----------------------------------------------------------------------------
  ! isobaric heat capacity (specific heat on constant pressure)
  !----------------------------------------------------------------------------

  function ThermoPropCp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    integer :: i


    Val = 0.0d0
    do i = 1, NCoef
      Val = Val + a_CPCoef(i) * ( ( Temp - Temp1 ) / Temp2 )**( i-1 )
    end do


  end function ThermoPropCp

  !----------------------------------------------------------------------------
  ! DPDRho
  !----------------------------------------------------------------------------

  function ThermoPropDPressDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Val = GasConst * Temp


  end function ThermoPropDPressDDensConstTemp

  !----------------------------------------------------------------------------
  ! DPDT
  !----------------------------------------------------------------------------

  function ThermoPropDPressDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    Val = Dens * GasConst


  end function ThermoPropDPressDTempConstDens

  !----------------------------------------------------------------------------
  ! DHDRho
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    Val = 0.0d0


  end function ThermoPropDEnthalpyDDensConstTemp

  !----------------------------------------------------------------------------
  ! DHDT
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    Val = ThermoPropCp( Dens, Temp )


  end function ThermoPropDEnthalpyDTempConstDens

  !----------------------------------------------------------------------------
  ! thermal expansion coefficient
  !----------------------------------------------------------------------------

  function ThermoPropThermoExpCoef( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !

    Val = 1.0d0 / Temp


  end function ThermoPropThermoExpCoef

  !----------------------------------------------------------------------------
  ! thermal expansion coefficient
  !----------------------------------------------------------------------------

  function ThermoPropAdLapRateDTDP( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Cp


    Cp = ThermoPropCp( Dens, Temp )

    Val = 1.0d0 / ( Cp * Dens )


  end function ThermoPropAdLapRateDTDP

  !----------------------------------------------------------------------------
  ! temperature and density from pressure and enthalpy
  !----------------------------------------------------------------------------

  subroutine ThermoPropDensTempFromPressEnthalpy( &
    & Press, Enthalpy, & ! (in)
    & Dens, Temp       & ! (out)
    & )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp), intent(out) :: Dens
    real(dp), intent(out) :: Temp


    ! Local variables
    !
    real(dp) :: EnthalpyCur
    real(dp) :: TempCur
    real(dp) :: TempNext
    real(dp) :: DHDTCRho
    real(dp) :: DelTemp
    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_Hist( 0:NLoop )

    integer  :: i


    ! Initialization
    TempCur = 300.0d0

    a_Hist(0) = TempCur
    search_temperature : do iLoop = 1, NLoop
      EnthalpyCur = 0.0d0
      do i = 1, NCoef
        EnthalpyCur = EnthalpyCur + a_CPCoef(i) / real( i ) * ( ( TempCur - Temp1 ) / Temp2 )**( i ) * Temp2
      end do
      DHDTCRho = 0.0d0
      do i = 1, NCoef
        DHDTCRho = DHDTCRho + a_CPCoef(i) * ( ( TempCur - Temp1 ) / Temp2 )**( i-1 )
      end do
      DelTemp = ( Enthalpy - EnthalpyCur ) / DHDTCRho
      TempNext = TempCur + DelTemp
      a_Hist( iLoop ) = TempNext
      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
        exit search_temperature
      end if
      TempCur  = TempNext
    end do search_temperature
    if ( iLoop >= NLoop ) then
      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Loop does not converge'
      write( 6, * ) 'ERROR:   ', iLoop
      write( 6, * ) 'ERROR:   ', a_Hist
      stop
    end if

    Temp = TempNext


    Dens = Press / ( GasConst * Temp )


  end subroutine ThermoPropDensTempFromPressEnthalpy

  !----------------------------------------------------------------------------

end module thermo_prop_co2_viracplsf
